from kurobako import problem
from kurobako.problem import Problem

from typing import List
from typing import Optional


class RosenbrockEvaluator(problem.Evaluator):
    """
    See https://www.sfu.ca/~ssurjano/rosen.html
    """

    def __init__(self, params: List[Optional[float]]):
        self._x1, self._x2 = params
        self._current_step = 0

    def evaluate(self, next_step: int) -> List[float]:
        self._current_step = 1
        value = 100 * (self._x2 - self._x1**2) ** 2 + (self._x1 - 1) ** 2
        return [value]

    def current_step(self) -> int:
        return self._current_step


class RosenbrockProblem(problem.Problem):
    def create_evaluator(
        self, params: List[Optional[float]]
    ) -> Optional[problem.Evaluator]:
        return RosenbrockEvaluator(params)


class RosenbrockProblemFactory(problem.ProblemFactory):
    def create_problem(self, seed: int) -> Problem:
        return RosenbrockProblem()

    def specification(self) -> problem.ProblemSpec:
        params = [
            problem.Var("x1", problem.ContinuousRange(-5, 10)),
            problem.Var("x2", problem.ContinuousRange(-5, 10)),
        ]
        return problem.ProblemSpec(
            name="Rosenbrock Function",
            params=params,
            values=[problem.Var("Rosenbrock")],
        )


if __name__ == "__main__":
    runner = problem.ProblemRunner(RosenbrockProblemFactory())
    runner.run()



import optuna


def objective(trial: optuna.Trial):
    x1 = trial.suggest_float("x1", -4, 4)
    x2 = trial.suggest_float("x2", -4, 4)
    return (x1 - 3) ** 2 + (10 * (x2 + 2)) ** 2


def main():
    optuna.logging.set_verbosity(optuna.logging.INFO)
    study = optuna.create_study(sampler=optuna.samplers.CmaEsSampler())
    study.optimize(objective, n_trials=250, gc_after_trial=False)


if __name__ == "__main__":
    main()



import sys
import numpy as np

from kurobako import problem
from kurobako.problem import Problem

from typing import List
from typing import Optional


class RastriginEvaluator(problem.Evaluator):
    def __init__(self, params: List[Optional[float]]):
        self.n = len(params)
        self.x = np.array(params, dtype=float)
        self._current_step = 0

    def evaluate(self, next_step: int) -> List[float]:
        self._current_step = 1
        value = 10 * self.n + np.sum(self.x**2 - 10 * np.cos(2 * np.pi * self.x))
        return [value]

    def current_step(self) -> int:
        return self._current_step


class RastriginProblem(problem.Problem):
    def create_evaluator(
        self, params: List[Optional[float]]
    ) -> Optional[problem.Evaluator]:
        return RastriginEvaluator(params)


class RastriginProblemFactory(problem.ProblemFactory):
    def __init__(self, dim):
        self.dim = dim

    def create_problem(self, seed: int) -> Problem:
        return RastriginProblem()

    def specification(self) -> problem.ProblemSpec:
        params = [
            problem.Var(f"x{i + 1}", problem.ContinuousRange(-5.12, 5.12))
            for i in range(self.dim)
        ]
        return problem.ProblemSpec(
            name=f"Rastrigin (dim={self.dim})",
            params=params,
            values=[problem.Var("Rastrigin")],
        )


if __name__ == "__main__":
    dim = int(sys.argv[1]) if len(sys.argv) == 2 else 2
    runner = problem.ProblemRunner(RastriginProblemFactory(dim))
    runner.run()



from kurobako import problem
from kurobako.problem import Problem

from typing import List
from typing import Optional


class HimmelblauEvaluator(problem.Evaluator):
    def __init__(self, params: List[Optional[float]]):
        self._x1, self._x2 = params
        self._current_step = 0

    def evaluate(self, next_step: int) -> List[float]:
        self._current_step = 1
        value = (self._x1**2 + self._x2 - 11.0) ** 2 + (
            self._x1 + self._x2**2 - 7.0
        ) ** 2
        return [value]

    def current_step(self) -> int:
        return self._current_step


class HimmelblauProblem(problem.Problem):
    def create_evaluator(
        self, params: List[Optional[float]]
    ) -> Optional[problem.Evaluator]:
        return HimmelblauEvaluator(params)


class HimmelblauProblemFactory(problem.ProblemFactory):
    def create_problem(self, seed: int) -> Problem:
        return HimmelblauProblem()

    def specification(self) -> problem.ProblemSpec:
        params = [
            problem.Var("x1", problem.ContinuousRange(-4, 4)),
            problem.Var("x2", problem.ContinuousRange(-4, 4)),
        ]
        return problem.ProblemSpec(
            name="Himmelblau Function",
            params=params,
            values=[problem.Var("Himmelblau")],
        )


if __name__ == "__main__":
    runner = problem.ProblemRunner(HimmelblauProblemFactory())
    runner.run()



import hypothesis.extra.numpy as npst
import unittest
from hypothesis import given, strategies as st

from cmaes import CMA, SepCMA


class TestFuzzing(unittest.TestCase):
    @given(
        data=st.data(),
    )
    def test_cma_tell(self, data):
        dim = data.draw(st.integers(min_value=2, max_value=100))
        mean = data.draw(npst.arrays(dtype=float, shape=dim))
        sigma = data.draw(st.floats(min_value=1e-16))
        n_iterations = data.draw(st.integers(min_value=1))
        try:
            optimizer = CMA(mean, sigma)
        except AssertionError:
            return
        popsize = optimizer.population_size
        for _ in range(n_iterations):
            tell_solutions = data.draw(
                st.lists(
                    st.tuples(npst.arrays(dtype=float, shape=dim), st.floats()),
                    min_size=popsize,
                    max_size=popsize,
                )
            )
            optimizer.ask()
            try:
                optimizer.tell(tell_solutions)
            except AssertionError:
                return
            optimizer.ask()

    @given(
        data=st.data(),
    )
    def test_sepcma_tell(self, data):
        dim = data.draw(st.integers(min_value=2, max_value=100))
        mean = data.draw(npst.arrays(dtype=float, shape=dim))
        sigma = data.draw(st.floats(min_value=1e-16))
        n_iterations = data.draw(st.integers(min_value=1))
        try:
            optimizer = SepCMA(mean, sigma)
        except AssertionError:
            return
        popsize = optimizer.population_size
        for _ in range(n_iterations):
            tell_solutions = data.draw(
                st.lists(
                    st.tuples(npst.arrays(dtype=float, shape=dim), st.floats()),
                    min_size=popsize,
                    max_size=popsize,
                )
            )
            optimizer.ask()
            try:
                optimizer.tell(tell_solutions)
            except AssertionError:
                return
            optimizer.ask()



import math
import numpy as np
from cmaes import CMA


def ackley(x1, x2):
    return (
        -20 * math.exp(-0.2 * math.sqrt(0.5 * (x1**2 + x2**2)))
        - math.exp(0.5 * (math.cos(2 * math.pi * x1) + math.cos(2 * math.pi * x2)))
        + math.e
        + 20
    )


def main():
    seed = 0
    rng = np.random.RandomState(1)

    bounds = np.array([[-32.768, 32.768], [-32.768, 32.768]])
    lower_bounds, upper_bounds = bounds[:, 0], bounds[:, 1]

    mean = lower_bounds + (rng.rand(2) * (upper_bounds - lower_bounds))
    sigma = 32.768 * 2 / 5  # 1/5 of the domain width
    optimizer = CMA(mean=mean, sigma=sigma, bounds=bounds, seed=0)

    # Multiplier for increasing population size before each restart.
    inc_popsize = 2

    print(" g    f(x1,x2)     x1      x2  ")
    print("===  ==========  ======  ======")
    for generation in range(200):
        solutions = []
        for _ in range(optimizer.population_size):
            x = optimizer.ask()
            value = ackley(x[0], x[1])
            solutions.append((x, value))
            print(f"{generation:3d}  {value:10.5f}  {x[0]:6.2f}  {x[1]:6.2f}")
        optimizer.tell(solutions)

        if optimizer.should_stop():
            seed += 1
            popsize = optimizer.population_size * inc_popsize
            mean = lower_bounds + (rng.rand(2) * (upper_bounds - lower_bounds))
            optimizer = CMA(
                mean=mean,
                sigma=sigma,
                bounds=bounds,
                seed=seed,
                population_size=popsize,
            )
            print("Restart CMA-ES with popsize={}".format(popsize))


if __name__ == "__main__":
    main()



import numpy as np
from cmaes import SepCMA


def ellipsoid(x):
    n = len(x)
    if len(x) < 2:
        raise ValueError("dimension must be greater one")
    return sum([(1000 ** (i / (n - 1)) * x[i]) ** 2 for i in range(n)])


def main():
    dim = 40
    optimizer = SepCMA(mean=3 * np.ones(dim), sigma=2.0)
    print(" evals    f(x)")
    print("======  ==========")

    evals = 0
    while True:
        solutions = []
        for _ in range(optimizer.population_size):
            x = optimizer.ask()
            value = ellipsoid(x)
            evals += 1
            solutions.append((x, value))
            if evals % 3000 == 0:
                print(f"{evals:5d}  {value:10.5f}")
        optimizer.tell(solutions)

        if optimizer.should_stop():
            break


if __name__ == "__main__":
    main()



import numpy as np
from cmaes import CMAwM


def ellipsoid_int(x, _):
    n = len(x)
    if len(x) < 2:
        raise ValueError("dimension must be greater one")
    return sum([(1000 ** (i / (n - 1)) * x[i]) ** 2 for i in range(n)])


def main():
    integer_dim, continuous_dim = 10, 10
    dim = integer_dim + continuous_dim
    bounds = np.concatenate(
        [
            np.tile([-np.inf, np.inf], (continuous_dim, 1)),
            np.tile([-10, 11], (integer_dim, 1)),
        ]
    )
    steps = np.concatenate([np.zeros(continuous_dim), np.ones(integer_dim)])
    optimizer = CMAwM(mean=5 * np.ones(dim), sigma=2.0, bounds=bounds, steps=steps)
    print(" evals    f(x)")
    print("======  ==========")

    evals = 0
    while True:
        solutions = []
        for _ in range(optimizer.population_size):
            x_for_eval, x_for_tell = optimizer.ask()
            value = ellipsoid_int(x_for_eval, integer_dim)
            evals += 1
            solutions.append((x_for_tell, value))
            if evals % 300 == 0:
                print(f"{evals:5d}  {value:10.5f}")
        optimizer.tell(solutions)

        if optimizer.should_stop():
            break


if __name__ == "__main__":
    main()



import warnings

import numpy as np
from numpy.testing import assert_almost_equal
from unittest import TestCase
from cmaes import CMA, CMAwM


class TestCMAwM(TestCase):
    def test_no_discrete_spaces(self):
        mean = np.zeros(2)
        bounds = np.array([[-10, 10], [-10, 10]])
        steps = np.array([0, 0])
        sigma = 1.3
        seed = 1

        cma_optimizer = CMA(mean=mean, sigma=sigma, bounds=bounds, seed=seed)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            cmawm_optimizer = CMAwM(
                mean=mean, sigma=sigma, bounds=bounds, steps=steps, seed=seed
            )

        for i in range(100):
            solutions = []
            for _ in range(cma_optimizer.population_size):
                cma_x = cma_optimizer.ask()
                cmawm_x_encoded, cmawm_x_for_tell = cmawm_optimizer.ask()
                assert_almost_equal(cma_x, cmawm_x_encoded)
                assert_almost_equal(cma_x, cmawm_x_for_tell)

                objective = (cma_x[0] - 3) ** 2 + cma_x[1] ** 2
                solutions.append((cma_x, objective))
            cma_optimizer.tell(solutions)
            cmawm_optimizer.tell(solutions)



import argparse
import optuna
import warnings

from kurobako import solver
from kurobako.solver.optuna import OptunaSolverFactory

warnings.filterwarnings(
    "ignore",
    category=optuna.exceptions.ExperimentalWarning,
    module="optuna.samplers._cmaes",
)

parser = argparse.ArgumentParser()
parser.add_argument(
    "sampler",
    choices=["cmaes", "sep-cmaes", "ipop-cmaes", "ipop-sep-cmaes", "pycma", "ws-cmaes"],
)
parser.add_argument(
    "--loglevel", choices=["debug", "info", "warning", "error"], default="warning"
)
parser.add_argument("--warm-starting-trials", type=int, default=0)
args = parser.parse_args()

if args.loglevel == "debug":
    optuna.logging.set_verbosity(optuna.logging.DEBUG)
elif args.loglevel == "info":
    optuna.logging.set_verbosity(optuna.logging.INFO)
elif args.loglevel == "warning":
    optuna.logging.set_verbosity(optuna.logging.WARNING)
elif args.loglevel == "error":
    optuna.logging.set_verbosity(optuna.logging.ERROR)


def create_cmaes_study(seed):
    sampler = optuna.samplers.CmaEsSampler(seed=seed, warn_independent_sampling=True)
    return optuna.create_study(sampler=sampler, pruner=optuna.pruners.NopPruner())


def create_sep_cmaes_study(seed):
    sampler = optuna.samplers.CmaEsSampler(
        seed=seed, warn_independent_sampling=True, use_separable_cma=True
    )
    return optuna.create_study(sampler=sampler, pruner=optuna.pruners.NopPruner())


def create_ipop_cmaes_study(seed):
    sampler = optuna.samplers.CmaEsSampler(
        seed=seed,
        warn_independent_sampling=True,
        restart_strategy="ipop",
        inc_popsize=2,
    )
    return optuna.create_study(sampler=sampler, pruner=optuna.pruners.NopPruner())


def create_ipop_sep_cmaes_study(seed):
    sampler = optuna.samplers.CmaEsSampler(
        seed=seed,
        warn_independent_sampling=True,
        restart_strategy="ipop",
        inc_popsize=2,
        use_separable_cma=True,
    )
    return optuna.create_study(sampler=sampler, pruner=optuna.pruners.NopPruner())


def create_pycma_study(seed):
    sampler = optuna.integration.PyCmaSampler(
        seed=seed,
        warn_independent_sampling=True,
    )
    return optuna.create_study(sampler=sampler, pruner=optuna.pruners.NopPruner())


class WarmStartingCmaEsSampler(optuna.samplers.BaseSampler):
    def __init__(self, seed, warm_starting_trials: int) -> None:
        self._seed = seed
        self._warm_starting = True
        self._warm_starting_trials = warm_starting_trials
        self._sampler = optuna.samplers.RandomSampler(seed=seed)
        self._source_trials = []

    def infer_relative_search_space(self, study, trial):
        return self._sampler.infer_relative_search_space(study, trial)

    def sample_relative(
        self,
        study,
        trial,
        search_space,
    ):
        return self._sampler.sample_relative(study, trial, search_space)

    def sample_independent(self, study, trial, param_name, param_distribution):
        return self._sampler.sample_independent(
            study, trial, param_name, param_distribution
        )

    def after_trial(
        self,
        study,
        trial,
        state,
        values,
    ):
        if not self._warm_starting:
            return self._sampler.after_trial(study, trial, state, values)

        if len(self._source_trials) < self._warm_starting_trials:
            assert state == optuna.trial.TrialState.PRUNED

            self._source_trials.append(
                optuna.create_trial(
                    params=trial.params,
                    distributions=trial.distributions,
                    values=values,
                )
            )
        if len(self._source_trials) == self._warm_starting_trials:
            self._sampler = optuna.samplers.CmaEsSampler(
                seed=self._seed + 1, source_trials=self._source_trials or None
            )
            self._warm_starting = False
        else:
            return self._sampler.after_trial(study, trial, state, values)


def create_warm_start_study(seed):
    sampler = WarmStartingCmaEsSampler(seed, args.warm_starting_trials)
    return optuna.create_study(sampler=sampler, pruner=optuna.pruners.NopPruner())


if __name__ == "__main__":
    if args.sampler == "cmaes":
        factory = OptunaSolverFactory(create_cmaes_study)
    elif args.sampler == "sep-cmaes":
        factory = OptunaSolverFactory(create_sep_cmaes_study)
    elif args.sampler == "ipop-cmaes":
        factory = OptunaSolverFactory(create_ipop_cmaes_study)
    elif args.sampler == "ipop-sep-cmaes":
        factory = OptunaSolverFactory(create_ipop_sep_cmaes_study)
    elif args.sampler == "pycma":
        factory = OptunaSolverFactory(create_pycma_study)
    elif args.sampler == "ws-cmaes":
        factory = OptunaSolverFactory(
            create_warm_start_study, warm_starting_trials=args.warm_starting_trials
        )
    else:
        raise ValueError("unsupported sampler")

    runner = solver.SolverRunner(factory)
    runner.run()



import math

from unittest import TestCase
from cmaes import _stats


# Test Cases in this file is generated by SciPy v1.9.3
class TestNormCDF(TestCase):
    def test_standard_normal_distribution(self):
        self.assertAlmostEqual(_stats.norm_cdf(-30), 4.906713927147907e-198, places=205)
        self.assertAlmostEqual(_stats.norm_cdf(-10), 7.619853024160469e-24, places=30)
        self.assertAlmostEqual(_stats.norm_cdf(-1), 0.15865525393145707)
        self.assertAlmostEqual(_stats.norm_cdf(0), 0.5)
        self.assertAlmostEqual(_stats.norm_cdf(1), 0.8413447460685429)
        self.assertAlmostEqual(
            _stats.norm_cdf(8),
            0.9999999999999993338661852249060757458209991455078125,
            places=30,
        )
        self.assertAlmostEqual(_stats.norm_cdf(10), 1.0)

    def test_mu_and_sigma(self):
        self.assertAlmostEqual(_stats.norm_cdf(1, loc=2, scale=3), 0.36944134018176367)


class TestChi2PPF(TestCase):
    def test(self):
        self.assertAlmostEqual(_stats.chi2_ppf(0.0), 0.0)
        self.assertAlmostEqual(
            _stats.chi2_ppf(0.00000001), 1.5707963267948962e-16, places=25
        )
        self.assertAlmostEqual(_stats.chi2_ppf(0.5), 0.454936423119572)
        self.assertAlmostEqual(_stats.chi2_ppf(0.99999999), 32.84125335146885)
        self.assertAlmostEqual(
            _stats.chi2_ppf(0.999999999999999777955395074969), 67.39648382445012
        )
        self.assertAlmostEqual(_stats.chi2_ppf(1.0), math.inf)



"""
Usage:
  cmaes_visualizer.py OPTIONS

Optional arguments:
  -h, --help            show this help message and exit
  --function {quadratic,himmelblau,rosenbrock,six-hump-camel}
  --seed SEED
  --frames FRAMES
  --interval INTERVAL
  --pop-per-frame POP_PER_FRAME
  --restart-strategy {ipop,bipop}

Example:
  python3 cmaes_visualizer.py --function six-hump-camel --pop-per-frame 2

  python3 tools/cmaes_visualizer.py --function himmelblau \
    --restart-strategy ipop --frames 500 --interval 10 --pop-per-frame 6
"""

import argparse
import math

import numpy as np
from scipy import stats

from matplotlib.colors import LinearSegmentedColormap
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from pylab import rcParams

from cmaes._cma import CMA

parser = argparse.ArgumentParser()
parser.add_argument(
    "--function",
    choices=["quadratic", "himmelblau", "rosenbrock", "six-hump-camel"],
)
parser.add_argument(
    "--seed",
    type=int,
    default=1,
)
parser.add_argument(
    "--frames",
    type=int,
    default=100,
)
parser.add_argument(
    "--interval",
    type=int,
    default=20,
)
parser.add_argument(
    "--pop-per-frame",
    type=int,
    default=1,
)
parser.add_argument(
    "--restart-strategy",
    choices=["ipop", "bipop"],
    default="",
)
args = parser.parse_args()

rcParams["figure.figsize"] = 10, 5
fig, (ax1, ax2) = plt.subplots(1, 2)

color_dict = {
    "red": ((0.0, 0.0, 0.0), (1.0, 1.0, 1.0)),
    "green": ((0.0, 0.0, 0.0), (1.0, 1.0, 1.0)),
    "blue": ((0.0, 1.0, 1.0), (1.0, 1.0, 1.0)),
    "yellow": ((1.0, 1.0, 1.0), (1.0, 1.0, 1.0)),
}
bw = LinearSegmentedColormap("BlueWhile", color_dict)


def himmelblau(x1, x2):
    return (x1**2 + x2 - 11.0) ** 2 + (x1 + x2**2 - 7.0) ** 2


def himmelblau_contour(x1, x2):
    return np.log(himmelblau(x1, x2) + 1)


def quadratic(x1, x2):
    return (x1 - 3) ** 2 + (10 * (x2 + 2)) ** 2


def quadratic_contour(x1, x2):
    return np.log(quadratic(x1, x2) + 1)


def rosenbrock(x1, x2):
    return 100 * (x2 - x1**2) ** 2 + (x1 - 1) ** 2


def rosenbrock_contour(x1, x2):
    return np.log(rosenbrock(x1, x2) + 1)


def six_hump_camel(x1, x2):
    return (
        (4 - 2.1 * (x1**2) + (x1**4) / 3) * (x1**2)
        + x1 * x2
        + (-4 + 4 * x2**2) * (x2**2)
    )


def six_hump_camel_contour(x1, x2):
    return np.log(six_hump_camel(x1, x2) + 1.0316)


function_name = ""
if args.function == "quadratic":
    function_name = "Quadratic function"
    objective = quadratic
    contour_function = quadratic_contour
    global_minimums = [
        (3.0, -2.0),
    ]
    # input domain
    x1_lower_bound, x1_upper_bound = -4, 4
    x2_lower_bound, x2_upper_bound = -4, 4
elif args.function == "himmelblau":
    function_name = "Himmelblau function"
    objective = himmelblau
    contour_function = himmelblau_contour
    global_minimums = [
        (3.0, 2.0),
        (-2.805118, 3.131312),
        (-3.779310, -3.283186),
        (3.584428, -1.848126),
    ]
    # input domain
    x1_lower_bound, x1_upper_bound = -4, 4
    x2_lower_bound, x2_upper_bound = -4, 4
elif args.function == "rosenbrock":
    # https://www.sfu.ca/~ssurjano/rosen.html
    function_name = "Rosenbrock function"
    objective = rosenbrock
    contour_function = rosenbrock_contour
    global_minimums = [
        (1, 1),
    ]
    # input domain
    x1_lower_bound, x1_upper_bound = -5, 10
    x2_lower_bound, x2_upper_bound = -5, 10
elif args.function == "six-hump-camel":
    # https://www.sfu.ca/~ssurjano/camel6.html
    function_name = "Six-hump camel function"
    objective = six_hump_camel
    contour_function = six_hump_camel_contour
    global_minimums = [
        (0.0898, -0.7126),
        (-0.0898, 0.7126),
    ]
    # input domain
    x1_lower_bound, x1_upper_bound = -3, 3
    x2_lower_bound, x2_upper_bound = -2, 2
else:
    raise ValueError("invalid function type")


seed = args.seed
bounds = np.array([[x1_lower_bound, x1_upper_bound], [x2_lower_bound, x2_upper_bound]])
sigma0 = (x1_upper_bound - x2_lower_bound) / 5
optimizer = CMA(mean=np.zeros(2), sigma=sigma0, bounds=bounds, seed=seed)
solutions = []
trial_number = 0
rng = np.random.RandomState(seed)

# Variables for IPOP and BIPOP
inc_popsize = 2
n_restarts = 0  # A small restart doesn't count in the n_restarts
small_n_eval, large_n_eval = 0, 0
popsize0 = optimizer.population_size
poptype = "small"


def init():
    ax1.set_xlim(x1_lower_bound, x1_upper_bound)
    ax1.set_ylim(x2_lower_bound, x2_upper_bound)
    ax2.set_xlim(x1_lower_bound, x1_upper_bound)
    ax2.set_ylim(x2_lower_bound, x2_upper_bound)

    # Plot 4 local minimum value
    for m in global_minimums:
        ax1.plot(m[0], m[1], "y*", ms=10)
        ax2.plot(m[0], m[1], "y*", ms=10)

    # Plot contour of himmelblau function
    x1 = np.arange(x1_lower_bound, x1_upper_bound, 0.01)
    x2 = np.arange(x2_lower_bound, x2_upper_bound, 0.01)
    x1, x2 = np.meshgrid(x1, x2)

    ax1.contour(x1, x2, contour_function(x1, x2), 30, cmap=bw)


def get_next_popsize_sigma():
    global optimizer, n_restarts, poptype, small_n_eval, large_n_eval, sigma0
    if args.restart_strategy == "ipop":
        n_restarts += 1
        popsize = optimizer.population_size * inc_popsize
        print(f"Restart CMA-ES with popsize={popsize} at trial={trial_number}")
        return popsize, sigma0
    elif args.restart_strategy == "bipop":
        n_eval = optimizer.population_size * optimizer.generation
        if poptype == "small":
            small_n_eval += n_eval
        else:  # poptype == "large"
            large_n_eval += n_eval

        if small_n_eval < large_n_eval:
            poptype = "small"
            popsize_multiplier = inc_popsize**n_restarts
            popsize = math.floor(popsize0 * popsize_multiplier ** (rng.uniform() ** 2))
            sigma = sigma0 * 10 ** (-2 * rng.uniform())
        else:
            poptype = "large"
            n_restarts += 1
            popsize = popsize0 * (inc_popsize**n_restarts)
            sigma = sigma0
        print(
            f"Restart CMA-ES with popsize={popsize} ({poptype}) at trial={trial_number}"
        )
        return popsize, sigma
    raise Exception("must not reach here")


def update(frame):
    global solutions, optimizer, trial_number
    if len(solutions) == optimizer.population_size:
        optimizer.tell(solutions)
        solutions = []

        if optimizer.should_stop():
            popsize, sigma = get_next_popsize_sigma()
            lower_bounds, upper_bounds = bounds[:, 0], bounds[:, 1]
            mean = lower_bounds + (rng.rand(2) * (upper_bounds - lower_bounds))
            optimizer = CMA(
                mean=mean,
                sigma=sigma,
                bounds=bounds,
                seed=seed,
                population_size=popsize,
            )

    n_sample = min(optimizer.population_size - len(solutions), args.pop_per_frame)
    for i in range(n_sample):
        x = optimizer.ask()
        evaluation = objective(x[0], x[1])

        # Plot sample points
        ax1.plot(x[0], x[1], "o", c="r", label="2d", alpha=0.5)

        solution = (
            x,
            evaluation,
        )
        solutions.append(solution)
    trial_number += n_sample

    # Update title
    if args.restart_strategy == "ipop":
        fig.suptitle(
            f"IPOP-CMA-ES {function_name} trial={trial_number} "
            f"popsize={optimizer.population_size}"
        )
    elif args.restart_strategy == "bipop":
        fig.suptitle(
            f"BIPOP-CMA-ES {function_name} trial={trial_number} "
            f"popsize={optimizer.population_size} ({poptype})"
        )
    else:
        fig.suptitle(f"CMA-ES {function_name} trial={trial_number}")

    # Plot multivariate gaussian distribution of CMA-ES
    x, y = np.mgrid[
        x1_lower_bound:x1_upper_bound:0.01, x2_lower_bound:x2_upper_bound:0.01
    ]
    rv = stats.multivariate_normal(optimizer._mean, optimizer._C)
    pos = np.dstack((x, y))
    ax2.contourf(x, y, rv.pdf(pos))

    if frame % 50 == 0:
        print(f"Processing frame {frame}")


def main():
    ani = animation.FuncAnimation(
        fig,
        update,
        frames=args.frames,
        init_func=init,
        blit=False,
        interval=args.interval,
    )
    ani.save(f"./tmp/{args.function}.mp4")


if __name__ == "__main__":
    main()



from kurobako import problem
from kurobako.problem import Problem

from typing import List
from typing import Optional


class SixHumpCamelEvaluator(problem.Evaluator):
    """
    See https://www.sfu.ca/~ssurjano/camel6.html
    """

    def __init__(self, params: List[Optional[float]]):
        self._x1, self._x2 = params
        self._current_step = 0

    def evaluate(self, next_step: int) -> List[float]:
        self._current_step = 1
        value = (
            (4 - 2.1 * (self._x1**2) + (self._x1**4) / 3) * (self._x1**2)
            + self._x1 * self._x2
            + (-4 + 4 * self._x2**2) * (self._x2**2)
        )
        return [value]

    def current_step(self) -> int:
        return self._current_step


class SixHumpCamelProblem(problem.Problem):
    def create_evaluator(
        self, params: List[Optional[float]]
    ) -> Optional[problem.Evaluator]:
        return SixHumpCamelEvaluator(params)


class SixHumpCamelProblemFactory(problem.ProblemFactory):
    def create_problem(self, seed: int) -> Problem:
        return SixHumpCamelProblem()

    def specification(self) -> problem.ProblemSpec:
        params = [
            problem.Var("x1", problem.ContinuousRange(-5, 10)),
            problem.Var("x2", problem.ContinuousRange(-5, 10)),
        ]
        return problem.ProblemSpec(
            name="Six-Hump Camel Function",
            params=params,
            values=[problem.Var("Six-Hump Camel")],
        )


if __name__ == "__main__":
    runner = problem.ProblemRunner(SixHumpCamelProblemFactory())
    runner.run()



import numpy as np
from cmaes import CMAwM


def ellipsoid_onemax(x, n_zdim):
    n = len(x)
    n_rdim = n - n_zdim
    r = 10
    if len(x) < 2:
        raise ValueError("dimension must be greater one")
    ellipsoid = sum([(1000 ** (i / (n_rdim - 1)) * x[i]) ** 2 for i in range(n_rdim)])
    onemax = n_zdim - (0.0 < x[(n - n_zdim) :]).sum()
    return ellipsoid + r * onemax


def main():
    binary_dim, continuous_dim = 10, 10
    dim = binary_dim + continuous_dim
    bounds = np.concatenate(
        [
            np.tile([-np.inf, np.inf], (continuous_dim, 1)),
            np.tile([0, 1], (binary_dim, 1)),
        ]
    )
    steps = np.concatenate([np.zeros(continuous_dim), np.ones(binary_dim)])
    optimizer = CMAwM(mean=np.zeros(dim), sigma=2.0, bounds=bounds, steps=steps)
    print(" evals    f(x)")
    print("======  ==========")

    evals = 0
    while True:
        solutions = []
        for _ in range(optimizer.population_size):
            x_for_eval, x_for_tell = optimizer.ask()
            value = ellipsoid_onemax(x_for_eval, binary_dim)
            evals += 1
            solutions.append((x_for_tell, value))
            if evals % 300 == 0:
                print(f"{evals:5d}  {value:10.5f}")
        optimizer.tell(solutions)

        if optimizer.should_stop():
            break


if __name__ == "__main__":
    main()



import numpy as np
from cmaes import CMA


def quadratic(x1, x2):
    return (x1 - 3) ** 2 + (10 * (x2 + 2)) ** 2


def main():
    optimizer = CMA(mean=np.zeros(2), sigma=1.3)
    print(" g    f(x1,x2)     x1      x2  ")
    print("===  ==========  ======  ======")

    while True:
        solutions = []
        for _ in range(optimizer.population_size):
            x = optimizer.ask()
            value = quadratic(x[0], x[1])
            solutions.append((x, value))
            print(
                f"{optimizer.generation:3d}  {value:10.5f}"
                f"  {x[0]:6.2f}  {x[1]:6.2f}"
            )
        optimizer.tell(solutions)

        if optimizer.should_stop():
            break


if __name__ == "__main__":
    main()



from __future__ import annotations

import math
import sys

import numpy as np

from typing import cast
from typing import Optional


_EPS = 1e-8
_MEAN_MAX = 1e32
_SIGMA_MAX = 1e32


class DXNESIC:
    """DX-NES-IC stochastic optimizer class with ask-and-tell interface.

    Example:

        .. code::

           import numpy as np
           from cmaes import DXNESIC

           def quadratic(x1, x2):
               return (x1 - 3) ** 2 + (10 * (x2 + 2)) ** 2

           optimizer = DXNESIC(mean=np.zeros(2), sigma=1.3)

           for generation in range(50):
               solutions = []
               for _ in range(optimizer.population_size):
                   # Ask a parameter
                   x = optimizer.ask()
                   value = quadratic(x[0], x[1])
                   solutions.append((x, value))
                   print(f"#{generation} {value} (x1={x[0]}, x2 = {x[1]})")

               # Tell evaluation values.
               optimizer.tell(solutions)

    Args:

        mean:
            Initial mean vector of multi-variate gaussian distributions.

        sigma:
            Initial standard deviation of covariance matrix.

        bounds:
            Lower and upper domain boundaries for each parameter (optional).

        n_max_resampling:
            A maximum number of resampling parameters (default: 100).
            If all sampled parameters are infeasible, the last sampled one
            will be clipped with lower and upper bounds.

        seed:
            A seed number (optional).

        population_size:
            A population size (optional).

        cov:
            A covariance matrix (optional).
    """

    # Paper: https://ieeexplore.ieee.org/abstract/document/9504865

    def __init__(
        self,
        mean: np.ndarray,
        sigma: float,
        bounds: Optional[np.ndarray] = None,
        n_max_resampling: int = 100,
        seed: Optional[int] = None,
        population_size: Optional[int] = None,
    ):
        assert sigma > 0, "sigma must be non-zero positive value"

        assert np.all(
            np.abs(mean) < _MEAN_MAX
        ), f"Abs of all elements of mean vector must be less than {_MEAN_MAX}"

        n_dim = len(mean)
        assert n_dim > 1, "The dimension of mean must be larger than 1"

        if population_size is None:
            population_size = 4 + math.floor(3 * math.log(n_dim))
        assert population_size > 0, "popsize must be non-zero positive value."

        w_rank_hat = np.log(population_size / 2 + 1) - np.log(
            np.arange(1, population_size + 1)
        )
        w_rank_hat[np.where(w_rank_hat < 0)] = 0
        w_rank = w_rank_hat / sum(w_rank_hat) - (1.0 / population_size)
        mu_eff = 1 / sum((w_rank + (1.0 / population_size)) ** 2)

        # learning rate for the cumulation for the step-size control
        c_sigma = (mu_eff + 2) / (n_dim + mu_eff + 5)
        assert (
            c_sigma < 1
        ), "invalid learning rate for cumulation for the step-size control"

        # distance weight parameter
        h_inv = _get_h_inv(n_dim)

        self._n_dim = n_dim
        self._popsize = population_size
        self._mu_eff = mu_eff

        self._h_inv = h_inv
        self._c_sigma = c_sigma

        # E||N(0, I)||
        self._chi_n = math.sqrt(self._n_dim) * (
            1.0 - (1.0 / (4.0 * self._n_dim)) + 1.0 / (21.0 * (self._n_dim**2))
        )

        # weights
        self._w_rank = w_rank
        self._w_rank_hat = w_rank_hat

        # for antithetic sampling
        self._zsym: Optional[np.ndarray] = None

        # learning rate
        self._eta_mean = 1.0
        self._eta_move_sigma = 1.0
        self._c_gamma = 1.0 / (3.0 * (n_dim - 1.0))
        self._d_gamma = min(1.0, n_dim / population_size)
        self._gamma = 1.0

        # evolution path
        self._p_sigma = np.zeros(n_dim)

        # distribution parameter
        self._mean = mean.copy()
        self._sigma = sigma
        self._B = np.eye(n_dim)

        # bounds contains low and high of each parameter.
        assert bounds is None or _is_valid_bounds(bounds, mean), "invalid bounds"
        self._bounds = bounds
        self._n_max_resampling = n_max_resampling

        self._g = 0
        self._rng = np.random.RandomState(seed)

        # Termination criteria
        self._tolx = 1e-12 * sigma
        self._tolxup = 1e4
        self._tolfun = 1e-12
        self._tolconditioncov = 1e14

        self._funhist_term = 10 + math.ceil(30 * n_dim / population_size)
        self._funhist_values = np.empty(self._funhist_term * 2)

    @property
    def dim(self) -> int:
        """A number of dimensions"""
        return self._n_dim

    @property
    def population_size(self) -> int:
        """A population size"""
        return self._popsize

    @property
    def generation(self) -> int:
        """Generation number which is monotonically incremented
        when multi-variate gaussian distribution is updated."""
        return self._g

    def _alpha_dist(self, num_feasible: int) -> float:
        return (
            self._h_inv
            * min(1.0, math.sqrt(float(self._popsize) / self._n_dim))
            * math.sqrt(float(num_feasible) / self._popsize)
        )

    def _w_dist_hat(self, z: np.ndarray, num_feasible: int) -> float:
        return math.exp(self._alpha_dist(num_feasible) * np.linalg.norm(z))

    def _eta_stag_sigma(self, num_feasible: int) -> float:
        return math.tanh(
            (0.024 * num_feasible + 0.7 * self._n_dim + 20.0) / (self._n_dim + 12.0)
        )

    def _eta_conv_sigma(self, num_feasible: int) -> float:
        return 2.0 * math.tanh(
            (0.025 * num_feasible + 0.75 * self._n_dim + 10.0) / (self._n_dim + 4.0)
        )

    def _eta_move_B(self, num_feasible: int) -> float:
        return (
            180
            * self._n_dim
            * math.tanh(0.02 * num_feasible)
            / (47 * (self._n_dim**2) + 6400)
        )

    def _eta_stag_B(self, num_feasible: int) -> float:
        return (
            168
            * self._n_dim
            * math.tanh(0.02 * num_feasible)
            / (47 * (self._n_dim**2) + 6400)
        )

    def _eta_conv_B(self, num_feasible: int) -> float:
        return (
            12
            * self._n_dim
            * math.tanh(0.02 * num_feasible)
            / (47 * (self._n_dim**2) + 6400)
        )

    def reseed_rng(self, seed: int) -> None:
        self._rng.seed(seed)

    def set_bounds(self, bounds: Optional[np.ndarray]) -> None:
        """Update boundary constraints"""
        assert bounds is None or _is_valid_bounds(bounds, self._mean), "invalid bounds"
        self._bounds = bounds

    def ask(self) -> np.ndarray:
        """Sample a parameter"""
        for i in range(self._n_max_resampling):
            x = self._sample_solution()
            if self._is_feasible(x):
                return x
        x = self._sample_solution()
        x = self._repair_infeasible_params(x)
        return x

    def _sample_solution(self) -> np.ndarray:
        # antithetic sampling
        if self._zsym is None:
            z = self._rng.randn(self._n_dim)  # ~ N(0, I)
            self._zsym = z
        else:
            z = -self._zsym
            self._zsym = None
        x = self._mean + self._sigma * self._B.dot(z)  # ~ N(m, σ^2 B B^T)
        return x

    def _is_feasible(self, param: np.ndarray) -> bool:
        if self._bounds is None:
            return True
        return cast(
            bool,
            np.all(param >= self._bounds[:, 0]) and np.all(param <= self._bounds[:, 1]),
        )  # Cast bool_ to bool.

    def _repair_infeasible_params(self, param: np.ndarray) -> np.ndarray:
        if self._bounds is None:
            return param

        # clip with lower and upper bound.
        param = np.where(param < self._bounds[:, 0], self._bounds[:, 0], param)
        param = np.where(param > self._bounds[:, 1], self._bounds[:, 1], param)
        return param

    def tell(self, solutions: list[tuple[np.ndarray, float]]) -> None:
        """Tell evaluation values"""

        assert len(solutions) == self._popsize, "Must tell popsize-length solutions."
        for s in solutions:
            assert np.all(
                np.abs(s[0]) < _MEAN_MAX
            ), f"Abs of all param values must be less than {_MEAN_MAX} to avoid overflow errors"

        # counting # feasible solutions
        lamb_feas = len([s[1] for s in solutions if s[1] < sys.maxsize])

        self._g += 1
        solutions.sort(key=lambda s: s[1])

        # Stores 'best' and 'worst' values of the
        # last 'self._funhist_term' generations.
        funhist_idx = 2 * (self.generation % self._funhist_term)
        self._funhist_values[funhist_idx] = solutions[0][1]
        self._funhist_values[funhist_idx + 1] = solutions[-1][1]

        z_k = np.array(
            [
                np.linalg.inv(self._sigma * self._B).dot(s[0] - self._mean)
                for s in solutions
            ]
        )

        # Evolution path
        z_w = np.sum(z_k.T * self._w_rank, axis=1)
        self._p_sigma = (1 - self._c_sigma) * self._p_sigma + math.sqrt(
            self._c_sigma * (2 - self._c_sigma) * self._mu_eff
        ) * z_w

        norm_p_sigma = np.linalg.norm(self._p_sigma)

        # switching learning rate depending on search situation
        movement_phase = norm_p_sigma >= self._chi_n

        # distance weight
        w_dist_tmp = np.array(
            [
                self._w_rank_hat[i] * self._w_dist_hat(z_k[i, :], lamb_feas)
                for i in range(self.population_size)
            ]
        )
        w_dist = w_dist_tmp / sum(w_dist_tmp) - 1.0 / self.population_size

        # switching weights and learning rate
        w = w_dist if movement_phase else self._w_rank
        eta_sigma = (
            self._eta_move_sigma
            if norm_p_sigma >= self._chi_n
            else (
                self._eta_stag_sigma(lamb_feas)
                if norm_p_sigma >= 0.1 * self._chi_n
                else self._eta_conv_sigma(lamb_feas)
            )
        )
        eta_B = (
            self._eta_move_B(lamb_feas)
            if norm_p_sigma >= self._chi_n
            else (
                self._eta_stag_B(lamb_feas)
                if norm_p_sigma >= 0.1 * self._chi_n
                else self._eta_conv_B(lamb_feas)
            )
        )

        # natural gradient estimation in local coordinate
        G_delta = np.sum(
            [w[i] * z_k[i, :] for i in range(self.population_size)], axis=0
        )
        G_M = np.sum(
            [
                w[i] * (np.outer(z_k[i, :], z_k[i, :]) - np.eye(self._n_dim))
                for i in range(self.population_size)
            ],
            axis=0,
        )
        G_sigma = G_M.trace() / self._n_dim
        G_B = G_M - G_sigma * np.eye(self._n_dim)

        # parameter update
        bBBT = self._B @ self._B.T
        self._mean += self._eta_mean * self._sigma * np.dot(self._B, G_delta)
        self._sigma *= math.exp((eta_sigma / 2.0) * G_sigma)
        # self._B = self._B.dot(expm((eta_B / 2.0) * G_B))
        self._B = self._B.dot(_expm((eta_B / 2.0) * G_B))
        aBBT = self._B @ self._B.T

        # emphasizing expansion
        e, v = np.linalg.eigh(bBBT)
        tau_vec = [
            (v[:, i].reshape(self._n_dim, 1).T @ aBBT @ v[:, i].reshape(self._n_dim, 1))
            / (
                v[:, i].reshape(self._n_dim, 1).T
                @ bBBT
                @ v[:, i].reshape(self._n_dim, 1)
            )
            - 1
            for i in range(self._n_dim)
        ]
        flg_tau = [1.0 if tau_vec[i] > 0 else 0.0 for i in range(self._n_dim)]
        tau = max(tau_vec)
        gamma = max(
            (1.0 - self._c_gamma) * self._gamma
            + self._c_gamma * math.sqrt(1.0 + self._d_gamma * tau),
            1.0,
        )
        if movement_phase:
            Q = (gamma - 1.0) * np.sum(
                [flg_tau[i] * np.outer(v[:, i], v[:, i]) for i in range(self._n_dim)],
                axis=0,
            ) + np.eye(self._n_dim)
            stepQ = math.pow(np.linalg.det(Q), 1.0 / self._n_dim)
            self._sigma *= stepQ
            self._B = Q @ self._B / stepQ

    def should_stop(self) -> bool:
        A = self._B.dot(self._B.T)
        A = (A + A.T) / 2
        E2, V = np.linalg.eigh(A)
        E = np.sqrt(np.where(E2 < 0, _EPS, E2))
        diagA = np.diag(A)

        # Stop if the range of function values of the recent generation is below tolfun.
        if (
            self.generation > self._funhist_term
            and np.max(self._funhist_values) - np.min(self._funhist_values)
            < self._tolfun
        ):
            return True

        # Stop if detecting divergent behavior.
        if self._sigma * np.max(E) > self._tolxup:
            return True

        # No effect coordinates: stop if adding 0.2-standard deviations
        # in any single coordinate does not change m.
        if np.any(self._mean == self._mean + (0.2 * self._sigma * np.sqrt(diagA))):
            return True

        # No effect axis: stop if adding 0.1-standard deviation vector in
        # any principal axis direction of C does not change m. "pycma" check
        # axis one by one at each generation.
        i = self.generation % self.dim
        if np.all(self._mean == self._mean + (0.1 * self._sigma * E[i] * V[:, i])):
            return True

        # Stop if the condition number of the covariance matrix exceeds 1e14.
        condition_cov = np.max(E) / np.min(E)
        if condition_cov > self._tolconditioncov:
            return True

        return False


def _is_valid_bounds(bounds: Optional[np.ndarray], mean: np.ndarray) -> bool:
    if bounds is None:
        return True
    if (mean.size, 2) != bounds.shape:
        return False
    if not np.all(bounds[:, 0] <= mean):
        return False
    if not np.all(mean <= bounds[:, 1]):
        return False
    return True


def _get_h_inv(dim: int) -> float:
    def f(a: float) -> float:
        return ((1.0 + a * a) * math.exp(a * a / 2.0) / 0.24) - 10.0 - dim

    def f_prime(a: float) -> float:
        return (1.0 / 0.24) * a * math.exp(a * a / 2.0) * (3.0 + a * a)

    h_inv = 6.0
    while abs(f(h_inv)) > 1e-10:
        last = h_inv
        h_inv = h_inv - 0.5 * (f(h_inv) / f_prime(h_inv))
        if abs(h_inv - last) < 1e-16:
            # Exit early since no further improvements are happening
            break
    return h_inv


def _expm(mat: np.ndarray) -> np.ndarray:
    D, U = np.linalg.eigh(mat)
    expD = np.exp(D)
    return U @ np.diag(expD) @ U.T



import numpy as np
from cmaes import CMA, get_warm_start_mgd


def source_task(x1: float, x2: float) -> float:
    b = 0.4
    return (x1 - b) ** 2 + (x2 - b) ** 2


def target_task(x1: float, x2: float) -> float:
    b = 0.6
    return (x1 - b) ** 2 + (x2 - b) ** 2


def main() -> None:
    # Generate solutions from a source task
    source_solutions = []
    for _ in range(1000):
        x = np.random.random(2)
        value = source_task(x[0], x[1])
        source_solutions.append((x, value))

    # Estimate a promising distribution of the source task
    ws_mean, ws_sigma, ws_cov = get_warm_start_mgd(
        source_solutions, gamma=0.1, alpha=0.1
    )
    optimizer = CMA(mean=ws_mean, sigma=ws_sigma, cov=ws_cov)

    # Run WS-CMA-ES
    print(" g    f(x1,x2)     x1      x2  ")
    print("===  ==========  ======  ======")
    while True:
        solutions = []
        for _ in range(optimizer.population_size):
            x = optimizer.ask()
            value = target_task(x[0], x[1])
            solutions.append((x, value))
            print(
                f"{optimizer.generation:3d}  {value:10.5f}"
                f"  {x[0]:6.2f}  {x[1]:6.2f}"
            )
        optimizer.tell(solutions)

        if optimizer.should_stop():
            break


if __name__ == "__main__":
    main()



import numpy as np
from cmaes import CMA


def rastrigin(x):
    dim = len(x)
    if dim < 2:
        raise ValueError("dimension must be greater one")
    return 10 * dim + sum(x**2 - 10 * np.cos(2 * np.pi * x))


if __name__ == "__main__":
    dim = 40
    optimizer = CMA(mean=3 * np.ones(dim), sigma=2.0, seed=10, lr_adapt=True)

    for generation in range(50000):
        solutions = []
        for _ in range(optimizer.population_size):
            x = optimizer.ask()
            value = rastrigin(x)
            if generation % 500 == 0:
                print(f"#{generation} {value}")
            solutions.append((x, value))
        optimizer.tell(solutions)

        if optimizer.should_stop():
            break



import numpy as np
from unittest import TestCase
from cmaes import CMA, get_warm_start_mgd


class TestWarmStartCMA(TestCase):
    def test_dimension(self):
        optimizer = CMA(mean=np.zeros(10), sigma=1.3)
        source_solutions = [(optimizer.ask(), 0.0) for _ in range(100)]
        ws_mean, ws_sigma, ws_cov = get_warm_start_mgd(source_solutions)

        self.assertEqual(ws_mean.size, 10)



import sys
import atheris
import hypothesis.extra.numpy as npst
from hypothesis import given, strategies as st

from cmaes import CMA


@given(data=st.data())
def test_cma_tell(data):
    dim = data.draw(st.integers(min_value=2, max_value=100))
    mean = data.draw(npst.arrays(dtype=float, shape=dim))
    sigma = data.draw(st.floats(min_value=1e-16))
    n_iterations = data.draw(st.integers(min_value=1))
    try:
        optimizer = CMA(mean, sigma)
    except AssertionError:
        return
    popsize = optimizer.population_size
    for _ in range(n_iterations):
        tell_solutions = data.draw(
            st.lists(
                st.tuples(npst.arrays(dtype=float, shape=dim), st.floats()),
                min_size=popsize,
                max_size=popsize,
            )
        )
        optimizer.ask()
        try:
            optimizer.tell(tell_solutions)
        except AssertionError:
            return
        optimizer.ask()


atheris.Setup(sys.argv, test_cma_tell.hypothesis.fuzz_one_input)
atheris.Fuzz()



from __future__ import annotations

import math
import numpy as np

from typing import Any
from typing import cast
from typing import Optional


_EPS = 1e-8
_MEAN_MAX = 1e32
_SIGMA_MAX = 1e32


class CatCMA:
    """CatCMA stochastic optimizer class with ask-and-tell interface.

    Example:

        .. code::

            import numpy as np
            from cmaes import CatCMA

            def sphere_com(x, c):
                return sum(x*x) + len(c) - sum(c[:,0])

            optimizer = CatCMA(mean=3 * np.ones(3), sigma=2.0, cat_num=np.array([3, 3, 3]))

            for generation in range(50):
                solutions = []
                for _ in range(optimizer.population_size):
                    # Ask a parameter
                    x, c = optimizer.ask()
                    value = sphere_com(x, c)
                    solutions.append(((x, c), value))
                    print(f"#{generation} {value}")

                # Tell evaluation values.
                optimizer.tell(solutions)

    Args:

        mean:
            Initial mean vector of multivariate gaussian distribution.

        sigma:
            Initial standard deviation of covariance matrix.

        cat_num:
            Numbers of categories.

        bounds:
            Lower and upper domain boundaries for each parameter (optional).

        n_max_resampling:
            A maximum number of resampling parameters (default: 100).
            If all sampled parameters are infeasible, the last sampled one
            will be clipped with lower and upper bounds.

        seed:
            A seed number (optional).

        population_size:
            A population size (optional).

        cov:
            A covariance matrix (optional).

        cat_param:
            A parameter of categorical distribution (optional).

        margin:
            A margin (lower bound) of categorical distribution (optional).

        min_eigenvalue:
            Lower bound of eigenvalue of multivariate Gaussian distribution (optional).
    """

    # Paper: https://arxiv.org/abs/2405.09962

    def __init__(
        self,
        mean: np.ndarray,
        sigma: float,
        cat_num: np.ndarray,
        bounds: Optional[np.ndarray] = None,
        n_max_resampling: int = 100,
        seed: Optional[int] = None,
        population_size: Optional[int] = None,
        cov: Optional[np.ndarray] = None,
        cat_param: Optional[np.ndarray] = None,
        margin: Optional[np.ndarray] = None,
        min_eigenvalue: Optional[float] = None,
    ):
        assert sigma > 0, "sigma must be non-zero positive value"

        assert np.all(
            np.abs(mean) < _MEAN_MAX
        ), f"Abs of all elements of mean vector must be less than {_MEAN_MAX}"

        self._n_co = len(mean)
        self._n_ca = len(cat_num)
        self._n = self._n_co + self._n_ca
        assert self._n_co > 1, "The dimension of mean must be larger than 1"
        assert self._n_ca > 0, "The dimension of categorical variable must be positive"
        assert np.all(cat_num > 1), "The number of categories must be larger than 1"

        if population_size is None:
            population_size = 4 + math.floor(3 * math.log(self._n))
        assert population_size > 0, "popsize must be non-zero positive value."

        mu = population_size // 2

        # CatCMA assumes that the weights of the lower half are zero.
        # (CMA uses negative weights while CatCMA uses positive weights.)
        weights_prime = np.array(
            [
                math.log((population_size + 1) / 2) - math.log(i + 1) if i < mu else 0
                for i in range(population_size)
            ]
        )
        weights = weights_prime / weights_prime.sum()
        mu_eff = 1 / ((weights**2).sum())

        # learning rate for the rank-one update
        alpha_cov = 2
        c1 = alpha_cov / ((self._n_co + 1.3) ** 2 + mu_eff)
        # learning rate for the rank-μ update
        cmu = min(
            1 - c1 - 1e-8,  # 1e-8 is for large popsize.
            alpha_cov
            * (mu_eff - 2 + 1 / mu_eff)
            / ((self._n_co + 2) ** 2 + alpha_cov * mu_eff / 2),
        )
        assert c1 <= 1 - cmu, "invalid learning rate for the rank-one update"
        assert cmu <= 1 - c1, "invalid learning rate for the rank-μ update"

        cm = 1

        # learning rate for the cumulation for the step-size control
        c_sigma = (mu_eff + 2) / (self._n_co + mu_eff + 5)
        d_sigma = (
            1 + 2 * max(0, math.sqrt((mu_eff - 1) / (self._n_co + 1)) - 1) + c_sigma
        )
        assert (
            c_sigma < 1
        ), "invalid learning rate for cumulation for the step-size control"

        # learning rate for cumulation for the rank-one update
        cc = (4 + mu_eff / self._n_co) / (self._n_co + 4 + 2 * mu_eff / self._n_co)
        assert cc <= 1, "invalid learning rate for cumulation for the rank-one update"

        self._popsize = population_size
        self._mu = mu
        self._mu_eff = mu_eff

        self._cc = cc
        self._c1 = c1
        self._cmu = cmu
        self._c_sigma = c_sigma
        self._d_sigma = d_sigma
        self._cm = cm

        # E||N(0, I)||
        self._chi_n = math.sqrt(self._n_co) * (
            1.0 - (1.0 / (4.0 * self._n_co)) + 1.0 / (21.0 * (self._n_co**2))
        )

        self._weights = weights

        # evolution path
        self._p_sigma = np.zeros(self._n_co)
        self._pc = np.zeros(self._n_co)

        self._mean = mean.copy()

        if cov is None:
            self._C = np.eye(self._n_co)
        else:
            assert cov.shape == (
                self._n_co,
                self._n_co,
            ), "Invalid shape of covariance matrix"
            self._C = cov

        self._sigma = sigma
        self._D: Optional[np.ndarray] = None
        self._B: Optional[np.ndarray] = None

        # categorical distribution
        # Parameters in categorical distribution with fewer categories
        # must be zero-padded at the end.
        self._K = cat_num
        self._Kmax = np.max(self._K)
        if cat_param is None:
            self._q = np.zeros((self._n_ca, self._Kmax))
            for i in range(self._n_ca):
                self._q[i, : self._K[i]] = 1 / self._K[i]
        else:
            assert cat_param.shape == (
                self._n_ca,
                self._Kmax,
            ), "Invalid shape of categorical distribution parameter"
            for i in range(self._n_ca):
                assert np.all(cat_param[i, self._K[i] :] == 0), (
                    "Parameters in categorical distribution with fewer categories "
                    "must be zero-padded at the end"
                )
            assert np.all(
                (cat_param >= 0) & (cat_param <= 1)
            ), "All elements in categorical distribution parameter must be between 0 and 1"
            assert np.allclose(
                np.sum(cat_param, axis=1), 1
            ), "Each row in categorical distribution parameter must sum to 1"
            self._q = cat_param

        self._q_min = (
            margin
            if margin is not None
            else (1 - 0.73 ** (1 / self._n_ca)) / (self._K - 1)
        )
        self._min_eigenvalue = min_eigenvalue if min_eigenvalue is not None else 1e-30

        # ASNG
        self._param_sum = np.sum(cat_num - 1)
        self._alpha = 1.5
        self._delta_init = 1.0
        self._Delta = 1.0
        self._Delta_max = np.inf
        self._gamma = 0.0
        self._s = np.zeros(self._param_sum)
        self._delta = self._delta_init / self._Delta
        self._eps = self._delta

        # bounds contains low and high of each parameter.
        assert bounds is None or _is_valid_bounds(bounds, mean), "invalid bounds"
        self._bounds = bounds
        self._n_max_resampling = n_max_resampling

        self._g = 0
        self._rng = np.random.RandomState(seed)

        # Termination criteria
        self._tolxup = 1e4
        self._tolfun = 1e-12
        self._tolconditioncov = 1e14

        self._funhist_term = 10 + math.ceil(30 * self._n_co / population_size)
        self._funhist_values = np.empty(self._funhist_term)

    def __getstate__(self) -> dict[str, Any]:
        attrs = {}
        for name in self.__dict__:
            # Remove _rng in pickle serialized object.
            if name == "_rng":
                continue
            if name == "_C":
                sym1d = _compress_symmetric(self._C)
                attrs["_c_1d"] = sym1d
                continue
            attrs[name] = getattr(self, name)
        return attrs

    def __setstate__(self, state: dict[str, Any]) -> None:
        state["_C"] = _decompress_symmetric(state["_c_1d"])
        del state["_c_1d"]
        self.__dict__.update(state)
        # Set _rng for unpickled object.
        setattr(self, "_rng", np.random.RandomState())

    @property
    def cont_dim(self) -> int:
        """A number of dimensions of continuous variable"""
        return self._n_co

    @property
    def cat_dim(self) -> int:
        """A number of dimensions of categorical variable"""
        return self._n_ca

    @property
    def dim(self) -> int:
        """A number of dimensions"""
        return self._n

    @property
    def cat_num(self) -> np.ndarray:
        """Numbers of categories"""
        return self._K

    @property
    def population_size(self) -> int:
        """A population size"""
        return self._popsize

    @property
    def generation(self) -> int:
        """Generation number which is monotonically incremented
        when multi-variate gaussian distribution is updated."""
        return self._g

    @property
    def mean(self) -> np.ndarray:
        """Mean Vector"""
        return self._mean

    def reseed_rng(self, seed: int) -> None:
        self._rng.seed(seed)

    def set_bounds(self, bounds: Optional[np.ndarray]) -> None:
        """Update boundary constraints"""
        assert bounds is None or _is_valid_bounds(bounds, self._mean), "invalid bounds"
        self._bounds = bounds

    def ask(self) -> tuple[np.ndarray, np.ndarray]:
        """Sample a parameter"""
        for i in range(self._n_max_resampling):
            x, c = self._sample_solution()
            if self._is_feasible(x):
                return x, c
        x, c = self._sample_solution()
        x = self._repair_infeasible_params(x)
        return x, c

    def _eigen_decomposition(self) -> tuple[np.ndarray, np.ndarray]:
        if self._B is not None and self._D is not None:
            return self._B, self._D

        self._C = (self._C + self._C.T) / 2
        D2, B = np.linalg.eigh(self._C)
        D = np.sqrt(np.where(D2 < 0, _EPS, D2))
        self._C = np.dot(np.dot(B, np.diag(D**2)), B.T)

        self._B, self._D = B, D
        return B, D

    def _sample_solution(self) -> tuple[np.ndarray, np.ndarray]:
        # x : continuous variable
        B, D = self._eigen_decomposition()
        z = self._rng.randn(self._n_co)  # ~ N(0, I)
        y = cast(np.ndarray, B.dot(np.diag(D))).dot(z)  # ~ N(0, C)
        x = self._mean + self._sigma * y  # ~ N(m, σ^2 C)
        # c : categorical variable
        # Categorical variables are one-hot encoded.
        # Variables with fewer categories are zero-padded at the end.
        rand_q = self._rng.rand(self._n_ca, 1)
        cum_q = self._q.cumsum(axis=1)
        c = (cum_q - self._q <= rand_q) & (rand_q < cum_q)
        return x, c

    def _is_feasible(self, param: np.ndarray) -> bool:
        if self._bounds is None:
            return True
        return cast(
            bool,
            np.all(param >= self._bounds[:, 0]) and np.all(param <= self._bounds[:, 1]),
        )  # Cast bool_ to bool.

    def _repair_infeasible_params(self, param: np.ndarray) -> np.ndarray:
        if self._bounds is None:
            return param

        # clip with lower and upper bound.
        param = np.where(param < self._bounds[:, 0], self._bounds[:, 0], param)
        param = np.where(param > self._bounds[:, 1], self._bounds[:, 1], param)
        return param

    def tell(
        self, solutions: list[tuple[tuple[np.ndarray, np.ndarray], float]]
    ) -> None:
        """Tell evaluation values"""

        assert len(solutions) == self._popsize, "Must tell popsize-length solutions."
        for s in solutions:
            assert np.all(
                np.abs(s[0][0]) < _MEAN_MAX
            ), f"Abs of all param values must be less than {_MEAN_MAX} to avoid overflow errors"

        self._g += 1
        solutions.sort(key=lambda s: s[1])

        # Stores best evaluation values of the
        # last 'self._funhist_term' generations.
        funhist_idx = self.generation % self._funhist_term
        self._funhist_values[funhist_idx] = solutions[0][1]

        # Sample new population of search_points, for k=1, ..., popsize
        B, D = self._eigen_decomposition()
        self._B, self._D = None, None

        x_k = np.array([s[0][0] for s in solutions])  # ~ N(m, σ^2 C)
        y_k = (x_k - self._mean) / self._sigma  # ~ N(0, C)

        # Selection and recombination
        y_w = np.sum(y_k[: self._mu].T * self._weights[: self._mu], axis=1)
        self._mean += self._cm * self._sigma * y_w

        # Step-size control
        C_2 = cast(
            np.ndarray, cast(np.ndarray, B.dot(np.diag(1 / D))).dot(B.T)
        )  # C^(-1/2) = B D^(-1) B^T
        self._p_sigma = (1 - self._c_sigma) * self._p_sigma + math.sqrt(
            self._c_sigma * (2 - self._c_sigma) * self._mu_eff
        ) * C_2.dot(y_w)

        norm_p_sigma = np.linalg.norm(self._p_sigma)
        self._sigma *= np.exp(
            (self._c_sigma / self._d_sigma) * (norm_p_sigma / self._chi_n - 1)
        )
        self._sigma = min(self._sigma, _SIGMA_MAX)

        # Covariance matrix adaption
        h_sigma_cond_left = norm_p_sigma / math.sqrt(
            1 - (1 - self._c_sigma) ** (2 * (self._g + 1))
        )
        h_sigma_cond_right = (1.4 + 2 / (self._n_co + 1)) * self._chi_n
        h_sigma = 1.0 if h_sigma_cond_left < h_sigma_cond_right else 0.0

        self._pc = (1 - self._cc) * self._pc + h_sigma * math.sqrt(
            self._cc * (2 - self._cc) * self._mu_eff
        ) * y_w

        delta_h_sigma = (1 - h_sigma) * self._cc * (2 - self._cc)
        assert delta_h_sigma <= 1

        rank_one = np.outer(self._pc, self._pc)
        rank_mu = np.sum(
            np.array([w * np.outer(y, y) for w, y in zip(self._weights, y_k)]), axis=0
        )
        self._C = (
            (
                1
                + self._c1 * delta_h_sigma
                - self._c1
                - self._cmu * np.sum(self._weights)
            )
            * self._C
            + self._c1 * rank_one
            + self._cmu * rank_mu
        )

        # Post-processing to prevent the minimum eigenvalue from becoming too small
        B, D = self._eigen_decomposition()
        sigma_min = np.sqrt(self._min_eigenvalue / np.min(D))
        self._sigma = max(self._sigma, sigma_min)

        # Update of categorical distribution
        c = np.array([s[0][1] for s in solutions])
        ngrad = (self._weights[:, np.newaxis, np.newaxis] * (c - self._q)).sum(axis=0)

        # Approximation of the square root of the fisher information matrix :
        # Appendix B in https://proceedings.mlr.press/v97/akimoto19a.html
        sl = []
        for i, K in enumerate(self._K):
            q_i = self._q[i, : K - 1]
            q_i_K = self._q[i, K - 1]
            s_i = 1.0 / np.sqrt(q_i) * ngrad[i, : K - 1]
            s_i += np.sqrt(q_i) * ngrad[i, : K - 1].sum() / (q_i_K + np.sqrt(q_i_K))
            sl += list(s_i)
        ngrad_sqF = np.array(sl)

        pnorm = np.sqrt(np.dot(ngrad_sqF, ngrad_sqF)) + 1e-30
        self._eps = self._delta / pnorm
        self._q += self._eps * ngrad

        # Update of ASNG
        self._delta = self._delta_init / self._Delta
        beta = self._delta / (self._param_sum**0.5)
        self._s = (1 - beta) * self._s + np.sqrt(beta * (2 - beta)) * ngrad_sqF / pnorm
        self._gamma = (1 - beta) ** 2 * self._gamma + beta * (2 - beta)
        self._Delta *= np.exp(
            beta * (self._gamma - np.dot(self._s, self._s) / self._alpha)
        )
        self._Delta = min(self._Delta, self._Delta_max)

        # Margin Correction
        for i in range(self._n_ca):
            Ki = self._K[i]
            self._q[i, :Ki] = np.maximum(self._q[i, :Ki], self._q_min[i])
            q_sum = self._q[i, :Ki].sum()
            tmp = q_sum - self._q_min[i] * Ki
            self._q[i, :Ki] -= (q_sum - 1) * (self._q[i, :Ki] - self._q_min[i]) / tmp
            self._q[i, :Ki] /= self._q[i, :Ki].sum()

    def should_stop(self) -> bool:
        B, D = self._eigen_decomposition()

        # Stop if the range of function values of the recent generation is below tolfun.
        if (
            self.generation > self._funhist_term
            and np.max(self._funhist_values) - np.min(self._funhist_values)
            < self._tolfun
        ):
            return True

        # Stop if detecting divergent behavior.
        if self._sigma * np.max(D) > self._tolxup:
            return True

        # Stop if the condition number of the covariance matrix exceeds 1e14.
        condition_cov = np.max(D) / np.min(D)
        if condition_cov > self._tolconditioncov:
            return True

        return False


def _is_valid_bounds(bounds: Optional[np.ndarray], mean: np.ndarray) -> bool:
    if bounds is None:
        return True
    if (mean.size, 2) != bounds.shape:
        return False
    if not np.all(bounds[:, 0] <= mean):
        return False
    if not np.all(mean <= bounds[:, 1]):
        return False
    return True


def _compress_symmetric(sym2d: np.ndarray) -> np.ndarray:
    assert len(sym2d.shape) == 2 and sym2d.shape[0] == sym2d.shape[1]
    n = sym2d.shape[0]
    dim = (n * (n + 1)) // 2
    sym1d = np.zeros(dim)
    start = 0
    for i in range(n):
        sym1d[start : start + n - i] = sym2d[i][i:]  # noqa: E203
        start += n - i
    return sym1d


def _decompress_symmetric(sym1d: np.ndarray) -> np.ndarray:
    n = int(np.sqrt(sym1d.size * 2))
    assert (n * (n + 1)) // 2 == sym1d.size
    R, C = np.triu_indices(n)
    out = np.zeros((n, n), dtype=sym1d.dtype)
    out[R, C] = sym1d
    out[C, R] = sym1d
    return out



import numpy as np
from unittest import TestCase
from cmaes._cma import _decompress_symmetric, _compress_symmetric


class TestCompressSymmetric(TestCase):
    def test_compress_symmetric_odd(self):
        sym2d = np.array([[1, 2], [2, 3]])
        actual = _compress_symmetric(sym2d)
        expected = np.array([1, 2, 3])
        self.assertTrue(np.all(np.equal(actual, expected)))

    def test_compress_symmetric_even(self):
        sym2d = np.array([[1, 2, 3], [2, 4, 5], [3, 5, 6]])
        actual = _compress_symmetric(sym2d)
        expected = np.array([1, 2, 3, 4, 5, 6])
        self.assertTrue(np.all(np.equal(actual, expected)))

    def test_decompress_symmetric_odd(self):
        sym1d = np.array([1, 2, 3])
        actual = _decompress_symmetric(sym1d)
        expected = np.array([[1, 2], [2, 3]])
        self.assertTrue(np.all(np.equal(actual, expected)))

    def test_decompress_symmetric_even(self):
        sym1d = np.array([1, 2, 3, 4, 5, 6])
        actual = _decompress_symmetric(sym1d)
        expected = np.array([[1, 2, 3], [2, 4, 5], [3, 5, 6]])
        self.assertTrue(np.all(np.equal(actual, expected)))



"""
Usage:
  python3 tools/ws_cmaes_visualizer.py OPTIONS

Optional arguments:
  -h, --help            show this help message and exit
  --function {quadratic,himmelblau,rosenbrock,six-hump-camel,sphere,rot-ellipsoid}
  --seed SEED
  --alpha ALPHA
  --gamma GAMMA
  --frames FRAMES
  --interval INTERVAL
  --pop-per-frame POP_PER_FRAME

Example:
  python3 ws_cmaes_visualizer.py --function rot-ellipsoid
"""

import argparse
import math

import numpy as np
from scipy import stats

from matplotlib.colors import LinearSegmentedColormap
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from pylab import rcParams

from cmaes import get_warm_start_mgd

parser = argparse.ArgumentParser()
parser.add_argument(
    "--function",
    choices=[
        "quadratic",
        "himmelblau",
        "rosenbrock",
        "six-hump-camel",
        "sphere",
        "rot-ellipsoid",
    ],
)
parser.add_argument(
    "--seed",
    type=int,
    default=1,
)
parser.add_argument(
    "--alpha",
    type=float,
    default=0.1,
)
parser.add_argument(
    "--gamma",
    type=float,
    default=0.1,
)
parser.add_argument(
    "--frames",
    type=int,
    default=100,
)
parser.add_argument(
    "--interval",
    type=int,
    default=20,
)
parser.add_argument(
    "--pop-per-frame",
    type=int,
    default=10,
)
args = parser.parse_args()

rcParams["figure.figsize"] = 10, 5
fig, (ax1, ax2) = plt.subplots(1, 2)

color_dict = {
    "red": ((0.0, 0.0, 0.0), (1.0, 1.0, 1.0)),
    "green": ((0.0, 0.0, 0.0), (1.0, 1.0, 1.0)),
    "blue": ((0.0, 1.0, 1.0), (1.0, 1.0, 1.0)),
    "yellow": ((1.0, 1.0, 1.0), (1.0, 1.0, 1.0)),
}
bw = LinearSegmentedColormap("BlueWhile", color_dict)


def himmelbleu(x1, x2):
    return (x1**2 + x2 - 11.0) ** 2 + (x1 + x2**2 - 7.0) ** 2


def himmelbleu_contour(x1, x2):
    return np.log(himmelbleu(x1, x2) + 1)


def quadratic(x1, x2):
    return (x1 - 3) ** 2 + (10 * (x2 + 2)) ** 2


def quadratic_contour(x1, x2):
    return np.log(quadratic(x1, x2) + 1)


def rosenbrock(x1, x2):
    return 100 * (x2 - x1**2) ** 2 + (x1 - 1) ** 2


def rosenbrock_contour(x1, x2):
    return np.log(rosenbrock(x1, x2) + 1)


def six_hump_camel(x1, x2):
    return (
        (4 - 2.1 * (x1**2) + (x1**4) / 3) * (x1**2)
        + x1 * x2
        + (-4 + 4 * x2**2) * (x2**2)
    )


def six_hump_camel_contour(x1, x2):
    return np.log(six_hump_camel(x1, x2) + 1.0316)


def sphere(x1, x2):
    offset = 0.6
    return (x1 - offset) ** 2 + (x2 - offset) ** 2


def sphere_contour(x1, x2):
    return np.log(sphere(x1, x2) + 1)


def ellipsoid(x1, x2):
    offset = 0.6
    scale = 5**2
    return (x1 - offset) ** 2 + scale * (x2 - offset) ** 2


def rot_ellipsoid(x1, x2):
    rot_x1 = math.sqrt(3.0) / 2.0 * x1 + 1.0 / 2.0 * x2
    rot_x2 = 1.0 / 2.0 * x1 + math.sqrt(3.0) / 2.0 * x2
    return ellipsoid(rot_x1, rot_x2)


def rot_ellipsoid_contour(x1, x2):
    return np.log(rot_ellipsoid(x1, x2) + 1)


function_name = ""
if args.function == "quadratic":
    function_name = "Quadratic function"
    objective = quadratic
    contour_function = quadratic_contour
    global_minimums = [
        (3.0, -2.0),
    ]
    # input domain
    x1_lower_bound, x1_upper_bound = -4, 4
    x2_lower_bound, x2_upper_bound = -4, 4
elif args.function == "himmelblau":
    function_name = "Himmelblau function"
    objective = himmelbleu
    contour_function = himmelbleu_contour
    global_minimums = [
        (3.0, 2.0),
        (-2.805118, 3.131312),
        (-3.779310, -3.283186),
        (3.584428, -1.848126),
    ]
    # input domain
    x1_lower_bound, x1_upper_bound = -4, 4
    x2_lower_bound, x2_upper_bound = -4, 4
elif args.function == "rosenbrock":
    # https://www.sfu.ca/~ssurjano/rosen.html
    function_name = "Rosenbrock function"
    objective = rosenbrock
    contour_function = rosenbrock_contour
    global_minimums = [
        (1, 1),
    ]
    # input domain
    x1_lower_bound, x1_upper_bound = -5, 10
    x2_lower_bound, x2_upper_bound = -5, 10
elif args.function == "six-hump-camel":
    # https://www.sfu.ca/~ssurjano/camel6.html
    function_name = "Six-hump camel function"
    objective = six_hump_camel
    contour_function = six_hump_camel_contour
    global_minimums = [
        (0.0898, -0.7126),
        (-0.0898, 0.7126),
    ]
    # input domain
    x1_lower_bound, x1_upper_bound = -3, 3
    x2_lower_bound, x2_upper_bound = -2, 2
elif args.function == "sphere":
    function_name = "Sphere function with offset=0.6"
    objective = sphere
    contour_function = sphere_contour
    global_minimums = [
        (0.6, 0.6),
    ]
    # input domain
    x1_lower_bound, x1_upper_bound = 0, 1
    x2_lower_bound, x2_upper_bound = 0, 1
elif args.function == "rot-ellipsoid":
    function_name = "Rot Ellipsoid function with offset=0.6"
    objective = rot_ellipsoid
    contour_function = rot_ellipsoid_contour

    global_minimums = []
    # input domain
    x1_lower_bound, x1_upper_bound = 0, 1
    x2_lower_bound, x2_upper_bound = 0, 1
else:
    raise ValueError("invalid function type")


seed = args.seed
rng = np.random.RandomState(seed)
solutions = []


def init():
    ax1.set_xlim(x1_lower_bound, x1_upper_bound)
    ax1.set_ylim(x2_lower_bound, x2_upper_bound)
    ax2.set_xlim(x1_lower_bound, x1_upper_bound)
    ax2.set_ylim(x2_lower_bound, x2_upper_bound)

    # Plot 4 local minimum value
    for m in global_minimums:
        ax1.plot(m[0], m[1], "y*", ms=10)
        ax2.plot(m[0], m[1], "y*", ms=10)

    # Plot contour of the function
    x1 = np.arange(x1_lower_bound, x1_upper_bound, 0.01)
    x2 = np.arange(x2_lower_bound, x2_upper_bound, 0.01)
    x1, x2 = np.meshgrid(x1, x2)

    ax1.contour(x1, x2, contour_function(x1, x2), 30, cmap=bw)


def update(frame):
    global solutions

    for i in range(args.pop_per_frame):
        x1 = (x1_upper_bound - x1_lower_bound) * rng.random() + x1_lower_bound
        x2 = (x2_upper_bound - x2_lower_bound) * rng.random() + x2_lower_bound

        evaluation = objective(x1, x2)

        # Plot sample points
        ax1.plot(x1, x2, "o", c="r", label="2d", alpha=0.5)

        solution = (
            np.array([x1, x2], dtype=float),
            evaluation,
        )
        solutions.append(solution)

    # Update title
    fig.suptitle(
        f"WS-CMA-ES {function_name} with alpha={args.alpha} and gamma={args.gamma} (frame={frame})"
    )

    # Plot multivariate gaussian distribution of CMA-ES
    x, y = np.mgrid[
        x1_lower_bound:x1_upper_bound:0.01, x2_lower_bound:x2_upper_bound:0.01
    ]

    if math.floor(len(solutions) * args.alpha) > 1:
        mean, sigma, cov = get_warm_start_mgd(
            solutions, alpha=args.alpha, gamma=args.gamma
        )
        rv = stats.multivariate_normal(mean, cov)
        pos = np.dstack((x, y))
        ax2.contourf(x, y, rv.pdf(pos))

    if frame % 50 == 0:
        print(f"Processing frame {frame}")


def main():
    ani = animation.FuncAnimation(
        fig,
        update,
        frames=args.frames,
        init_func=init,
        blit=False,
        interval=args.interval,
    )
    ani.save(f"./tmp/{args.function}.mp4")


if __name__ == "__main__":
    main()



import argparse
import cProfile
import logging
import pstats
import optuna

parser = argparse.ArgumentParser()
parser.add_argument("--storage", choices=["memory", "sqlite"], default="memory")
parser.add_argument("--params", type=int, default=100)
parser.add_argument("--trials", type=int, default=1000)
args = parser.parse_args()


def objective(trial: optuna.Trial):
    val = 0
    for i in range(args.params):
        xi = trial.suggest_uniform(str(i), -4, 4)
        val += (xi - 2) ** 2
    return val


def main():
    logging.disable(level=logging.INFO)
    storage = None
    if args.storage == "sqlite":
        storage = f"sqlite:///db-{args.trials}-{args.params}.sqlite3"
    sampler = optuna.samplers.CmaEsSampler()
    study = optuna.create_study(sampler=sampler, storage=storage)

    profiler = cProfile.Profile()
    profiler.runcall(
        study.optimize, objective, n_trials=args.trials, gc_after_trial=False
    )
    profiler.dump_stats("profile.stats")

    stats = pstats.Stats("profile.stats")
    stats.sort_stats("time").print_stats(5)


if __name__ == "__main__":
    main()



import numpy as np
from cmaes import CatCMA


def sphere_com(x, c):
    dim_co = len(x)
    dim_ca = len(c)
    if dim_co < 2:
        raise ValueError("dimension must be greater one")
    sphere = sum(x * x)
    com = dim_ca - sum(c[:, 0])
    return sphere + com


def rosenbrock_clo(x, c):
    dim_co = len(x)
    dim_ca = len(c)
    if dim_co < 2:
        raise ValueError("dimension must be greater one")
    rosenbrock = sum(100 * (x[:-1] ** 2 - x[1:]) ** 2 + (x[:-1] - 1) ** 2)
    clo = dim_ca - (c[:, 0].argmin() + c[:, 0].prod() * dim_ca)
    return rosenbrock + clo


def mc_proximity(x, c, cat_num):
    dim_co = len(x)
    dim_ca = len(c)
    if dim_co < 2:
        raise ValueError("dimension must be greater one")
    if dim_co != dim_ca:
        raise ValueError(
            "number of dimensions of continuous and categorical variables "
            "must be equal in mc_proximity"
        )

    c_index = np.argmax(c, axis=1) / cat_num
    return sum((x - c_index) ** 2) + sum(c_index)


if __name__ == "__main__":
    cont_dim = 5
    cat_dim = 5
    cat_num = np.array([3, 4, 5, 5, 5])
    # cat_num = 3 * np.ones(cat_dim, dtype=np.int64)
    optimizer = CatCMA(mean=3.0 * np.ones(cont_dim), sigma=1.0, cat_num=cat_num)

    for generation in range(200):
        solutions = []
        for _ in range(optimizer.population_size):
            x, c = optimizer.ask()
            value = mc_proximity(x, c, cat_num)
            if generation % 10 == 0:
                print(f"#{generation} {value}")
            solutions.append(((x, c), value))
        optimizer.tell(solutions)

        if optimizer.should_stop():
            break



import math
import numpy as np
from cmaes import CMA


def ackley(x1, x2):
    return (
        -20 * math.exp(-0.2 * math.sqrt(0.5 * (x1**2 + x2**2)))
        - math.exp(0.5 * (math.cos(2 * math.pi * x1) + math.cos(2 * math.pi * x2)))
        + math.e
        + 20
    )


def main():
    seed = 0
    rng = np.random.RandomState(0)

    bounds = np.array([[-32.768, 32.768], [-32.768, 32.768]])
    lower_bounds, upper_bounds = bounds[:, 0], bounds[:, 1]

    mean = lower_bounds + (rng.rand(2) * (upper_bounds - lower_bounds))
    sigma0 = 32.768 * 2 / 5  # 1/5 of the domain width
    sigma = sigma0
    optimizer = CMA(mean=mean, sigma=sigma, bounds=bounds, seed=0)

    n_restarts = 0  # A small restart doesn't count in the n_restarts
    small_n_eval, large_n_eval = 0, 0
    popsize0 = optimizer.population_size
    inc_popsize = 2

    # Initial run is with "normal" population size; it is
    # the large population before first doubling, but its
    # budget accounting is the same as in case of small
    # population.
    poptype = "small"

    while n_restarts <= 5:
        solutions = []
        for _ in range(optimizer.population_size):
            x = optimizer.ask()
            value = ackley(x[0], x[1])
            solutions.append((x, value))
            # print("{:10.5f}  {:6.2f}  {:6.2f}".format(value, x[0], x[1]))
        optimizer.tell(solutions)

        if optimizer.should_stop():
            seed += 1
            n_eval = optimizer.population_size * optimizer.generation
            if poptype == "small":
                small_n_eval += n_eval
            else:  # poptype == "large"
                large_n_eval += n_eval

            if small_n_eval < large_n_eval:
                poptype = "small"
                popsize_multiplier = inc_popsize**n_restarts
                popsize = math.floor(
                    popsize0 * popsize_multiplier ** (rng.uniform() ** 2)
                )
                sigma = sigma0 * 10 ** (-2 * rng.uniform())
            else:
                poptype = "large"
                n_restarts += 1
                popsize = popsize0 * (inc_popsize**n_restarts)
                sigma = sigma0
            mean = lower_bounds + (rng.rand(2) * (upper_bounds - lower_bounds))
            optimizer = CMA(
                mean=mean,
                sigma=sigma,
                bounds=bounds,
                seed=seed,
                population_size=popsize,
            )
            print("Restart CMA-ES with popsize={} ({})".format(popsize, poptype))


if __name__ == "__main__":
    main()



from ._cma import CMA  # NOQA
from ._sepcma import SepCMA  # NOQA
from ._warm_start import get_warm_start_mgd  # NOQA
from ._cmawm import CMAwM  # NOQA
from ._xnes import XNES  # NOQA
from ._dxnesic import DXNESIC  # NOQA
from ._catcma import CatCMA  # NOQA

__version__ = "0.11.1"



import numpy as np
from unittest import TestCase
from cmaes import CMA, SepCMA


CMA_CLASSES = [CMA, SepCMA]


class TestCMABoundary(TestCase):
    def test_valid_dimension(self):
        for CmaClass in CMA_CLASSES:
            with self.subTest(f"Class: {CmaClass.__name__}"):
                CmaClass(
                    mean=np.zeros(2), sigma=1.3, bounds=np.array([[-10, 10], [-10, 10]])
                )

    def test_invalid_dimension(self):
        for CmaClass in CMA_CLASSES:
            with self.subTest(f"Class: {CmaClass.__name__}"):
                with self.assertRaises(AssertionError):
                    CmaClass(mean=np.zeros(2), sigma=1.3, bounds=np.array([-10, 10]))

    def test_mean_located_out_of_bounds(self):
        mean = np.zeros(5)
        bounds = np.empty(shape=(5, 2))
        bounds[:, 0], bounds[:, 1] = 1.0, 5.0
        for CmaClass in CMA_CLASSES:
            with self.subTest(f"Class: {CmaClass.__name__}"):
                with self.assertRaises(AssertionError):
                    CmaClass(mean=mean, sigma=1.3, bounds=bounds)

    def test_set_valid_bounds(self):
        for CmaClass in CMA_CLASSES:
            with self.subTest(f"Class: {CmaClass.__name__}"):
                optimizer = CmaClass(mean=np.zeros(2), sigma=1.3)
                optimizer.set_bounds(bounds=np.array([[-10, 10], [-10, 10]]))

    def test_set_invalid_bounds(self):
        for CmaClass in CMA_CLASSES:
            with self.subTest(f"Class: {CmaClass.__name__}"):
                optimizer = CmaClass(mean=np.zeros(2), sigma=1.3)
                with self.assertRaises(AssertionError):
                    optimizer.set_bounds(bounds=np.array([-10, 10]))

    def test_set_bounds_which_does_not_contain_mean(self):
        for CmaClass in CMA_CLASSES:
            with self.subTest(f"Class: {CmaClass.__name__}"):
                optimizer = CmaClass(mean=np.zeros(2), sigma=1.3)
                bounds = np.empty(shape=(5, 2))
                bounds[:, 0], bounds[:, 1] = 1.0, 5.0
                with self.assertRaises(AssertionError):
                    optimizer.set_bounds(bounds)



import numpy as np

from unittest import TestCase
from cmaes import CMA


class TestTerminationCriterion(TestCase):
    def test_stop_if_objective_values_are_not_changed(self):
        optimizer = CMA(mean=np.zeros(2), sigma=1.3)
        popsize = optimizer.population_size
        rng = np.random.RandomState(seed=1)

        for i in range(optimizer._funhist_term + 1):
            self.assertFalse(optimizer.should_stop())
            optimizer.tell([(rng.randn(2), 0.01) for _ in range(popsize)])

        self.assertTrue(optimizer.should_stop())

    def test_stop_if_detect_divergent_behavior(self):
        optimizer = CMA(mean=np.zeros(2), sigma=1e-4)
        popsize = optimizer.population_size
        nd_rng = np.random.RandomState(1)

        solutions = [(100 * nd_rng.randn(2), 0.01) for _ in range(popsize)]
        optimizer.tell(solutions)
        self.assertTrue(optimizer.should_stop())
